from typing import List, Optional

import structlog

from rasa.core.agent import Agent
from rasa.core.channels import UserMessage
from rasa.dialogue_understanding.commands import Command, SetSlotCommand
from rasa.llm_fine_tuning.conversations import Conversation, ConversationStep
from rasa.llm_fine_tuning.paraphrasing.rephrased_user_message import (
    RephrasedUserMessage,
)
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.utils.llm import (
    create_tracker_for_user_step,
    generate_sender_id,
)

structlogger = structlog.get_logger()


class RephraseValidator:
    def __init__(self, flows: FlowsList) -> None:
        self.flows = flows

    async def validate_rephrasings(
        self,
        agent: Agent,
        rephrasings: List[RephrasedUserMessage],
        conversation: Conversation,
    ) -> List[RephrasedUserMessage]:
        """Split rephrased user messages into passing and failing.

        Handle the rephrased messages using agent the same way the original
        message was handled. Check if the rephrased user message is producing
        the same commands as the original user message. The rephrase is passing
        if the commands match and failing otherwise.

        Args:
            agent: Rasa agent
            rephrasings: The rephrased user messages.
            conversation: The conversation.

        Returns:
            A list of rephrased user messages including the passing and failing
            rephrases.
        """
        for i, step in enumerate(
            conversation.iterate_over_annotated_user_steps(rephrase=True)
        ):
            current_rephrasings = rephrasings[i]

            for rephrase in current_rephrasings.rephrasings:
                if await self._validate_rephrase_is_passing(
                    agent,
                    rephrase,
                    step,
                    conversation.name,
                    conversation.tracker,
                ):
                    current_rephrasings.passed_rephrasings.append(rephrase)
                else:
                    current_rephrasings.failed_rephrasings.append(rephrase)

        return rephrasings

    async def _validate_rephrase_is_passing(
        self,
        agent: Agent,
        rephrase: str,
        step: ConversationStep,
        test_case_name: str,
        tracker: DialogueStateTracker,
    ) -> bool:
        rephrased_tracker = await self._send_rephrased_message_to_agent(
            rephrase, step, test_case_name, agent, tracker
        )
        if not (rephrased_tracker and rephrased_tracker.latest_message):
            return False

        commands_from_original_utterance = step.llm_commands

        commands_from_rephrased_utterance = [
            Command.command_from_json(command_json)
            for command_json in rephrased_tracker.latest_message.commands
        ]

        return self._check_commands_match(
            commands_from_original_utterance, commands_from_rephrased_utterance
        )

    @staticmethod
    def _check_commands_match(
        expected_commands: List[Command], actual_commands: List[Command]
    ) -> bool:
        if len(expected_commands) != len(actual_commands):
            return False

        for expected_command in expected_commands:
            if isinstance(expected_command, SetSlotCommand):
                slot_name = expected_command.name
                match_found = False
                for c in actual_commands:
                    if isinstance(c, SetSlotCommand) and c.name == slot_name:
                        match_found = True
                        break
                if not match_found:
                    return False

            elif expected_command not in actual_commands:
                return False

        return True

    @staticmethod
    async def _send_rephrased_message_to_agent(
        rephrased_user_message: str,
        step: ConversationStep,
        test_case_name: str,
        agent: Agent,
        tracker: DialogueStateTracker,
    ) -> Optional[DialogueStateTracker]:
        # create a rephrased UserMessage
        sender_id = generate_sender_id(test_case_name)
        user_message = UserMessage(rephrased_user_message, sender_id=sender_id)

        await create_tracker_for_user_step(
            sender_id, agent, tracker, step.tracker_event_index
        )

        await agent.handle_message(user_message)
        rephrased_tracker = await agent.tracker_store.retrieve(sender_id)

        return rephrased_tracker



"""Copilot execution utilities for evaluators.

This module provides utilities for running copilot operations in evaluation contexts,
independent of specific evaluation frameworks like Langfuse.
"""

from typing import List, Optional

import structlog
from pydantic import BaseModel

from rasa.builder.config import COPILOT_HANDLER_ROLLING_BUFFER_SIZE
from rasa.builder.copilot.models import (
    CopilotContext,
    CopilotGenerationContext,
    GeneratedContent,
    ReferenceSection,
    ResponseCategory,
)
from rasa.builder.llm_service import llm_service

structlogger = structlog.get_logger()


class CopilotRunResult(BaseModel):
    """Result from running the copilot with response handler."""

    complete_response: Optional[str]
    response_category: Optional[ResponseCategory]
    reference_section: Optional[ReferenceSection]
    generation_context: CopilotGenerationContext


async def run_copilot_with_response_handler(
    context: CopilotContext,
) -> Optional[CopilotRunResult]:
    """Run the copilot with response handler on the given context.

    This function encapsulates the core copilot execution logic. It handles:
    - Instantiating the copilot and response handler
    - Generating a response and extracting the reference section from the given context
    - Returning structured results

    Args:
        context: The copilot context to process.

    Returns:
        CopilotRunResult containing the complete response, category, and generation
        context, or None if execution fails.

    Raises:
        Any exceptions from the copilot or response handler execution.
    """
    # Instantiate the copilot and response handler
    copilot = llm_service.instantiate_copilot()
    copilot_response_handler = llm_service.instantiate_handler(
        COPILOT_HANDLER_ROLLING_BUFFER_SIZE
    )

    # Call the copilot to generate a response and handle it with the response
    # handler
    (original_stream, generation_context) = await copilot.generate_response(context)
    intercepted_stream = copilot_response_handler.handle_response(original_stream)

    # Exhaust the stream to get the complete response for evaluation
    response_chunks: List[str] = []
    response_category = None
    async for chunk in intercepted_stream:
        if not isinstance(chunk, GeneratedContent):
            continue
        response_chunks.append(chunk.content)
        response_category = chunk.response_category

    complete_response = "".join(response_chunks) if response_chunks else None

    # Extract the reference section from the response handler
    if generation_context.relevant_documents:
        reference_section = copilot_response_handler.extract_references(
            generation_context.relevant_documents
        )
    else:
        reference_section = None

    return CopilotRunResult(
        complete_response=complete_response,
        response_category=response_category,
        reference_section=reference_section,
        generation_context=generation_context,
    )



from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, Union

from rasa.shared.utils.yaml import read_config_file


class CredentialsConfig:
    def __init__(
        self, channels: Dict[str, Dict[str, Any]], config_file_path: Path
    ) -> None:
        self.channels = channels
        self.config_file_path = config_file_path

        for _, config in self.channels.items():
            if config and "silence_timeout" in config:
                config["silence_timeout"] = self.validate_silence_timeout(
                    config["silence_timeout"]
                )

    @classmethod
    def load_from_file(cls, file_path: Path) -> CredentialsConfig:
        credentials_dict = read_config_file(file_path)
        return cls(credentials_dict, file_path)

    @classmethod
    def validate_silence_timeout(
        cls, silence_timeout: Union[float, int]
    ) -> Union[float, int]:
        """Validate the silence timeout value."""
        if isinstance(silence_timeout, str):
            try:
                silence_timeout = float(silence_timeout)
            except ValueError:
                raise ValueError(
                    f"Type error for silence timeout value: {silence_timeout}. "
                    "Silence timeout must be a positive number."
                )

        if not isinstance(silence_timeout, (float, int)):
            raise ValueError(
                f"Type error for silence timeout value: {silence_timeout}. "
                "Silence timeout must be a positive number."
            )

        if silence_timeout <= 0:
            raise ValueError(
                f"Value error for silence timeout value: {silence_timeout}. "
                "Silence timeout must be a positive number."
            )

        return silence_timeout



import logging
from urllib.parse import urlparse

import structlog
from sanic import Sanic

from rasa.cli.scaffold import print_error_and_exit
from rasa.core.persistor import RemoteStorageType, get_persistor
from rasa.core.utils import list_routes
from rasa.model_manager import config, model_api
from rasa.model_manager.config import SERVER_BASE_URL, SERVER_PORT
from rasa.utils.common import configure_logging_and_warnings
from rasa.utils.log_utils import configure_structlog
from rasa.utils.sanic_error_handler import register_custom_sanic_error_handler

structlogger = structlog.get_logger()


def url_prefix_from_base_url() -> str:
    """Return the path prefix from the base URL."""
    # return path without any trailing slashes
    prefix = urlparse(SERVER_BASE_URL).path.rstrip("/") if SERVER_BASE_URL else ""

    # can't be empty
    return prefix or config.DEFAULT_SERVER_PATH_PREFIX


def validate_model_storage_type() -> None:
    """Validate the storage type if remote storage is used for models."""
    if config.SERVER_MODEL_REMOTE_STORAGE:
        if config.SERVER_MODEL_REMOTE_STORAGE not in RemoteStorageType.list():
            print_error_and_exit(
                f"Invalid storage type '{config.SERVER_MODEL_REMOTE_STORAGE}'. "
                f"Supported storage types: {', '.join(RemoteStorageType.list())}."
                f"Alternatively, unset the remote storage type to store models locally."
            )
        else:
            structlogger.info(
                "model_api.storage.remote_storage_enabled",
                remote_storage=config.SERVER_MODEL_REMOTE_STORAGE,
            )
        # try to create a client to validate the configuration
        get_persistor(config.SERVER_MODEL_REMOTE_STORAGE)
    else:
        structlogger.info(
            "model_api.storage.local_storage_enabled",
            base_path=config.SERVER_BASE_WORKING_DIRECTORY,
        )


def _register_update_task(app: Sanic) -> None:
    app.add_task(
        model_api.continuously_update_process_status,
        name="continuously_update_process_status",
    )


def main() -> None:
    """Start the Rasa Model Manager server.

    The API server can receive requests to train models, run bots, and manage
    the lifecycle of models and bots.
    """
    import rasa.utils.licensing

    log_level = logging.DEBUG
    configure_logging_and_warnings(
        log_level=log_level,
        logging_config_file=None,
        warn_only_once=True,
        filter_repeated_logs=True,
    )
    configure_structlog(log_level, include_time=True)

    rasa.utils.licensing.validate_license_from_env()

    try:
        model_api.prepare_working_directories()
    except Exception as e:
        structlogger.error(
            "model_api.prepare_directories.failed",
            error=str(e),
            base_directory=config.SERVER_BASE_WORKING_DIRECTORY,
        )
        print_error_and_exit(
            f"Failed to create working directories. Please make sure the "
            f"server base directory at '{config.SERVER_BASE_WORKING_DIRECTORY}' "
            f"is writable by the current user."
        )

    validate_model_storage_type()

    structlogger.debug("model_api.starting_server", port=SERVER_PORT)

    url_prefix = url_prefix_from_base_url()
    # configure the sanic application
    app = Sanic("RasaModelService")
    app.after_server_start(_register_update_task)
    app.blueprint(model_api.external_blueprint(), url_prefix=url_prefix)
    app.blueprint(model_api.internal_blueprint())

    # list all routes
    list_routes(app)

    register_custom_sanic_error_handler(app)

    app.run(host="0.0.0.0", port=SERVER_PORT, legacy=True, motd=False)


if __name__ == "__main__":
    main()



from typing import Any, Dict, List, Text

from rasa_sdk import Action, Tracker
from rasa_sdk.events import SlotSet
from rasa_sdk.executor import CollectingDispatcher

from actions.db import get_account


class CheckTransferLimit(Action):
    def name(self) -> Text:
        return "check_transfer_limit"

    def run(
        self,
        dispatcher: CollectingDispatcher,
        tracker: Tracker,
        domain: Dict[Text, Any],
    ) -> List[Dict[Text, Any]]:
        limit_type = tracker.get_slot("transfer_limit_type")

        account = get_account(tracker.sender_id)
        daily_limit = account.daily_limit
        monthly_limit = account.monthly_limit

        transfer_limit = None

        if limit_type == "daily":
            transfer_limit = daily_limit
        elif limit_type == "monthly":
            transfer_limit = monthly_limit

        return [
            SlotSet("transfer_limit", transfer_limit),
            SlotSet("return_value", transfer_limit is not None),
        ]



import audioop
import wave
from dataclasses import asdict, dataclass
from typing import Dict, Optional, Type, TypeVar

import structlog

from rasa.core.channels.voice_stream.audio_bytes import HERTZ, RasaAudioBytes
from rasa.shared.exceptions import RasaException

structlogger = structlog.get_logger()


def read_wav_to_rasa_audio_bytes(file_name: str) -> Optional[RasaAudioBytes]:
    """Reads rasa audio bytes from a file."""
    if not file_name.endswith(".wav"):
        raise RasaException("Should only read .wav files with this method.")
    wave_object = wave.open(file_name, "rb")
    wave_data = wave_object.readframes(wave_object.getnframes())
    if wave_object.getnchannels() != 1:
        wave_data = audioop.tomono(wave_data, wave_object.getsampwidth(), 1, 1)
    if wave_object.getsampwidth() != 1:
        wave_data = audioop.lin2lin(wave_data, wave_object.getsampwidth(), 1)
        # 8 bit is unsigned
        # wave_data = audioop.bias(wave_data, 1, 128)
    if wave_object.getframerate() != HERTZ:
        wave_data, _ = audioop.ratecv(
            wave_data, 1, 1, wave_object.getframerate(), HERTZ, None
        )
    wave_data = audioop.lin2ulaw(wave_data, 1)
    return RasaAudioBytes(wave_data)


def generate_silence(length_in_seconds: float = 1.0) -> RasaAudioBytes:
    return RasaAudioBytes(b"\00" * int(length_in_seconds * HERTZ))


T = TypeVar("T", bound="MergeableConfig")


@dataclass
class MergeableConfig:
    def __init__(self) -> None:
        pass

    def merge(self: T, other: Optional[T]) -> T:
        """Merges two configs while dropping None values of the second config."""
        if other is None:
            return self
        other_dict = asdict(other)
        other_dict_clean = {k: v for k, v in other_dict.items() if v is not None}
        merged = {**asdict(self), **other_dict_clean}
        return self.from_dict(merged)

    @classmethod
    def from_dict(cls: Type[T], data: dict[str, Optional[str]]) -> T:
        return cls(**data)


def repack_voice_credentials(
    credentials: Dict[str, str],
) -> Dict[str, str]:
    """Repack voice credentials to ensure they are in the correct format."""
    new_creds = {**credentials}
    new_creds["asr_config"] = new_creds.pop("asr", None)
    new_creds["tts_config"] = new_creds.pop("tts", None)
    return new_creds



from typing import Any, Dict


class Singleton(type):
    """Singleton metaclass."""

    _instances: Dict[Any, Any] = {}  # noqa: RUF012

    def __call__(cls, *args: Any, **kwargs: Any) -> Any:
        """Call the class.

        Args:
            *args: Arguments.
            **kwargs: Keyword arguments.
        """
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)

        return cls._instances[cls]

    def clear(cls) -> None:
        """Clear the class."""
        cls._instances = {}
